@@ -17,9 +17,9 @@
  *   sudo make install
  * 
  * Last modified:
- *   30/08/2016
+ *   14/01/2019
  *
- * Copyright 2016, Ondrej Wisniewski 
+ * Copyright 2015-2019, Ondrej Wisniewski 
  * 
  * This file is part of crelay.
  * 
@@ -57,12 +57,12 @@
 #include "config.h"
 #include "relay_drv.h"
 
-#define VERSION "0.10.1"
-#define DATE "2016"
+#define VERSION "0.14"
+#define DATE "2019"
 
 /* HTTP server defines */
 #define SERVER "crelay/"VERSION
-#define PROTOCOL "HTTP/1.0"
+#define PROTOCOL "HTTP/1.1"
 #define RFC1123FMT "%a, %d %b %Y %H:%M:%S GMT"
 #define API_URL "gpio"
 #define DEFAULT_SERVER_PORT 8000
@@ -70,6 +70,7 @@
 /* HTML tag definitions */
 #define RELAY_TAG "pin"
 #define STATE_TAG "status"
+#define SERIAL_TAG "serial"
 
 #define CONFIG_FILE "/etc/crelay.conf"
 
@@ -79,7 +80,6 @@ config_t config;
 static char rlabels[MAX_NUM_RELAYS][32] = {"My appliance 1", "My appliance 2", "My appliance 3", "My appliance 4",
                                            "My appliance 5", "My appliance 6", "My appliance 7", "My appliance 8"};                                       
 
-
 /**********************************************************
  * Function: config_cb()
  * 
@@ -134,10 +134,18 @@ static int config_cb(void* user, const char* section, const char* name, const ch
    {
       pconfig->relay8_label = strdup(value);
    } 
+   else if (MATCH("HTTP server", "pulse_duration")) 
+   {
+      pconfig->pulse_duration = atoi(value);
+   }
    else if (MATCH("GPIO drv", "num_relays")) 
    {
       pconfig->gpio_num_relays = atoi(value);
    } 
+   else if (MATCH("GPIO drv", "active_value")) 
+   {
+      pconfig->gpio_active_value = atoi(value);
+   } 
    else if (MATCH("GPIO drv", "relay1_gpio_pin")) 
    {
       pconfig->relay1_gpio_pin = atoi(value);
@@ -215,11 +223,12 @@ void send_headers(FILE *f, int status, char *title, char *extra, char *mime,
    
    fprintf(f, "%s %d %s\r\n", PROTOCOL, status, title);
    fprintf(f, "Server: %s\r\n", SERVER);
+   //fprintf(f, "Access-Control-Allow-Origin: *\r\n"); // TEST For test only
    now = time(NULL);
    strftime(timebuf, sizeof(timebuf), RFC1123FMT, gmtime(&now));
    fprintf(f, "Date: %s\r\n", timebuf);
    if (extra) fprintf(f, "%s\r\n", extra);
-   if (mime) fprintf(f, "Content-Type: %s\r\n", mime);
+   if (mime) fprintf(f, "Content-Type: %s; charset=utf-8\r\n", mime);
    if (length >= 0) fprintf(f, "Content-Length: %d\r\n", length);
    if (date != -1)
    {
@@ -231,6 +240,103 @@ void send_headers(FILE *f, int status, char *title, char *extra, char *mime,
 }
 
 
+/**********************************************************
+ * Function java_script_src()
+ * 
+ * Description:
+ * 
+ * Parameters:
+ * 
+ *********************************************************/
+void java_script_src(FILE *f)
+{
+   fprintf(f, "<script type='text/javascript'>\r\n");
+   fprintf(f, "function switch_relay(checkboxElem){\r\n");
+   fprintf(f, "   var status = checkboxElem.checked ? 1 : 0;\r\n");
+   fprintf(f, "   var pin = checkboxElem.id;\r\n");
+   fprintf(f, "   var url = '/gpio?pin='+pin+'&status='+status;\r\n");
+   fprintf(f, "   var xmlHttp = new XMLHttpRequest();\r\n");
+   fprintf(f, "   xmlHttp.onreadystatechange = function () {\r\n");
+   fprintf(f, "      if (this.readyState < 4)\r\n");
+   fprintf(f, "         document.getElementById('status').innerHTML = '';\r\n");
+   fprintf(f, "      else if (this.readyState == 4) {\r\n"); 
+   fprintf(f, "         if (this.status == 0) {\r\n");
+   fprintf(f, "            document.getElementById('status').innerHTML = \"Network error\";\r\n");
+   fprintf(f, "            checkboxElem.checked = (status==0);\r\n");
+   fprintf(f, "         }\r\n");
+   fprintf(f, "         else if (this.status != 200) {\r\n");
+   fprintf(f, "            document.getElementById('status').innerHTML = this.statusText;\r\n");
+   fprintf(f, "            checkboxElem.checked = (status==0);\r\n");
+   fprintf(f, "         }\r\n");
+   // TODO: add update of all relays status here (according to API response in xmlHttp.responseText) 
+   fprintf(f, "      }\r\n");
+   fprintf(f, "   }\r\n");
+   fprintf(f, "   xmlHttp.open( 'GET', url, true );\r\n");
+   fprintf(f, "   xmlHttp.send( null );\r\n");
+   fprintf(f, "}\r\n");
+   fprintf(f, "</script>\r\n");
+}
+
+
+/**********************************************************
+ * Function style_sheet()
+ * 
+ * Description:
+ * 
+ * Parameters:
+ * 
+ *********************************************************/
+void style_sheet(FILE *f)
+{
+   fprintf(f, "<style>\r\n");
+   fprintf(f, ".switch {\r\n");
+   fprintf(f, "  position: relative;\r\n");
+   fprintf(f, "  display: inline-block;\r\n");
+   fprintf(f, "  width: 60px;\r\n");
+   fprintf(f, "  height: 34px;\r\n");
+   fprintf(f, "}\r\n");
+   fprintf(f, ".switch input {\r\n"); 
+   fprintf(f, "  opacity: 0;\r\n");
+   fprintf(f, "  width: 0;\r\n");
+   fprintf(f, "  height: 0;\r\n");
+   fprintf(f, "}\r\n");
+   fprintf(f, ".slider {\r\n");
+   fprintf(f, "  position: absolute;\r\n");
+   fprintf(f, "  cursor: pointer;\r\n");
+   fprintf(f, "  top: 0;\r\n");
+   fprintf(f, "  left: 0;\r\n");
+   fprintf(f, "  right: 0;\r\n");
+   fprintf(f, "  bottom: 0;\r\n");
+   fprintf(f, "  background-color: #ccc;\r\n");
+   fprintf(f, "  -webkit-transition: .4s;\r\n");
+   fprintf(f, "  transition: .4s;\r\n");
+   fprintf(f, "}\r\n");
+   fprintf(f, ".slider:before {\r\n");
+   fprintf(f, "  position: absolute;\r\n");
+   fprintf(f, "  content: \"\";\r\n");
+   fprintf(f, "  height: 26px;\r\n");
+   fprintf(f, "  width: 26px;\r\n");
+   fprintf(f, "  left: 4px;\r\n");
+   fprintf(f, "  bottom: 4px;\r\n");
+   fprintf(f, "  background-color: white;\r\n");
+   fprintf(f, "  -webkit-transition: .4s;\r\n");
+   fprintf(f, "  transition: .4s;\r\n");
+   fprintf(f, "}\r\n");
+   fprintf(f, "input:checked + .slider {\r\n");
+   fprintf(f, "  background-color: #2196F3;\r\n");
+   fprintf(f, "}\r\n");
+   fprintf(f, "input:focus + .slider {\r\n");
+   fprintf(f, "  box-shadow: 0 0 1px #2196F3;\r\n");
+   fprintf(f, "}\r\n");
+   fprintf(f, "input:checked + .slider:before {\r\n");
+   fprintf(f, "  -webkit-transform: translateX(26px);\r\n");
+   fprintf(f, "  -ms-transform: translateX(26px);\r\n");
+   fprintf(f, "  transform: translateX(26px);\r\n");
+   fprintf(f, "}\r\n");
+   fprintf(f, "</style>\r\n");   
+}
+
+
 /**********************************************************
  * Function web_page_header()
  * 
@@ -242,16 +348,19 @@ void send_headers(FILE *f, int status, char *title, char *extra, char *mime,
 void web_page_header(FILE *f)
 {
    /* Send http header */
-   send_headers(f, 200, "OK", NULL, "text/html", -1, -1);
+   send_headers(f, 200, "OK", NULL, "text/html", -1, -1);   
+   fprintf(f, "<!DOCTYPE html PUBLIC \"-//W3C//DTD HTML 4.01//EN\" \"http://www.w3.org/TR/html4/strict.dtd\">\r\n");
+   fprintf(f, "<html><head><title>Relay Card Control</title>\r\n");
+   style_sheet(f);
+   java_script_src(f);
+   fprintf(f, "</head>\r\n");
    
    /* Display web page heading */
-   fprintf(f, "<!DOCTYPE html PUBLIC \"-//W3C//DTD HTML 4.01//EN\" \"http://www.w3.org/TR/html4/strict.dtd\">\r\n");
-   fprintf(f, "<html><head><title>Relay Card Control interface</title></head><body>\r\n");
-   fprintf(f, "<table style=\"text-align: left; width: 600px; background-color: rgb(0, 0, 153); font-family: Helvetica,Arial,sans-serif; font-weight: bold; color: white;\" border=\"0\" cellpadding=\"2\" cellspacing=\"2\">\r\n");
+   fprintf(f, "<body><table style=\"text-align: left; width: 460px; background-color: #2196F3; font-family: Helvetica,Arial,sans-serif; font-weight: bold; color: white;\" border=\"0\" cellpadding=\"2\" cellspacing=\"2\">\r\n");
    fprintf(f, "<tbody><tr><td>\r\n");
    fprintf(f, "<span style=\"vertical-align: top; font-size: 48px;\">Relay Card Control</span><br>\r\n");
-   fprintf(f, "<span style=\"font-size: 16px; color: rgb(204, 255, 255);\">Remote relay card controlling <span style=\"font-style: italic; color: white;\">made easy</span></span>\r\n");
-   fprintf(f, "</td></tr></tbody></table>\r\n");  
+   fprintf(f, "<span style=\"font-size: 16px; color: rgb(204, 255, 255);\">Remote relay card control <span style=\"font-style: italic; color: white;\">made easy</span></span>\r\n");
+   fprintf(f, "</td></tr></tbody></table><br>\r\n");  
 }
 
 
@@ -266,8 +375,8 @@ void web_page_header(FILE *f)
 void web_page_footer(FILE *f)
 {
    /* Display web page footer */
-   fprintf(f, "<table style=\"text-align: left; width: 600px; background-color: rgb(0, 0, 153);\" border=\"0\" cellpadding=\"2\" cellspacing=\"2\"><tbody>\r\n");
-   fprintf(f, "<tr><td style=\"vertical-align: top; text-align: center;\"><span style=\"font-family: Helvetica,Arial,sans-serif; color: white;\">crelay | version %s | %s</span></td></tr>\r\n",
+   fprintf(f, "<table style=\"text-align: left; width: 460px; background-color: #2196F3;\" border=\"0\" cellpadding=\"2\" cellspacing=\"2\"><tbody>\r\n");
+   fprintf(f, "<tr><td style=\"vertical-align: top; text-align: center;\"><span style=\"font-family: Helvetica,Arial,sans-serif; color: white;\"><a style=\"text-decoration:none; color: white;\" href=http://ondrej1024.github.io/crelay>crelay</a> | version %s | %s</span></td></tr>\r\n",
            VERSION, DATE);
    fprintf(f, "</tbody></table></body></html>\r\n");
 }   
@@ -284,7 +393,7 @@ void web_page_footer(FILE *f)
 void web_page_error(FILE *f)
 {    
    /* No relay card detected, display error message on web page */
-   fprintf(f, "<br><table style=\"text-align: left; width: 600px; background-color: yellow; font-family: Helvetica,Arial,sans-serif; font-weight: bold; color: black;\" border=\"0\" cellpadding=\"2\" cellspacing=\"2\">\r\n");
+   fprintf(f, "<br><table style=\"text-align: left; width: 460px; background-color: yellow; font-family: Helvetica,Arial,sans-serif; font-weight: bold; color: black;\" border=\"0\" cellpadding=\"2\" cellspacing=\"2\">\r\n");
    fprintf(f, "<tbody><tr style=\"font-size: 20px; font-weight: bold;\">\r\n");
    fprintf(f, "<td>No compatible relay card detected !<br>\r\n");
    fprintf(f, "<span style=\"font-size: 14px; color: grey;  font-weight: normal;\">This can be due to the following reasons:\r\n");
@@ -304,7 +413,7 @@ void web_page_error(FILE *f)
  * Parameters:
  * 
  *********************************************************/
-int read_httppost_data(FILE* f, char* data)
+int read_httppost_data(FILE* f, char* data, size_t datalen)
 {
    char buf[256];
    int data_len=0;
@@ -313,6 +422,7 @@ int read_httppost_data(FILE* f, char* data)
     * Therefore we skip the header and then read the form
     * data into the buffer.
     */
+   *data = 0;
    while (fgets(buf, sizeof(buf), f) != NULL)
    {
       //printf("%s", buf);
@@ -326,7 +436,11 @@ int read_httppost_data(FILE* f, char* data)
       /* Find end of header (empty line) */
       if (!strcmp(buf, "\r\n")) break;
    }
-   
+
+   /* Make sure we're not trying to overwrite the buffer */
+   if (data_len >= datalen)
+     return -1;
+
    /* Get form data string */
    if (!fgets(data, data_len+1, f)) return -1;
    *(data+data_len) = 0;
@@ -343,18 +457,20 @@ int read_httppost_data(FILE* f, char* data)
  * Parameters:
  * 
  *********************************************************/
-int read_httpget_data(char* buf, char* data)
+int read_httpget_data(char* buf, char* data, size_t datalen)
 {
    char *datastr;
          
    /* GET request: data (if any) is provided in the first line
     * of the page header. Therefore we first check if there is
     * any data. If so we read the data into a buffer.
+    *
+    * Note that we may truncate the input if it's too long.
     */
    *data = 0;
    if ((datastr=strchr(buf, '?')) != NULL)
    {
-      strcpy(data, datastr+1);
+       strncpy(data, datastr+1, datalen);
    }
    
    return strlen(data);
@@ -379,8 +495,10 @@ int process_http_request(int sock)
    char *datastr;
    int  relay=0;
    int  i;
+   int  formdatalen;
    char com_port[MAX_COM_PORT_NAME_LEN];
-   uint8 last_relay=FIRST_RELAY;
+   char* serial=NULL;
+   uint8_t last_relay=FIRST_RELAY;
    relay_state_t rstate[MAX_NUM_RELAYS];
    relay_state_t nstate=INVALID;
    
@@ -419,11 +537,11 @@ int process_http_request(int sock)
    /* Check the request method we are dealing with */
    if (strcasecmp(method, "POST") == 0)
    {
-      read_httppost_data(fin, formdata);
+      formdatalen = read_httppost_data(fin, formdata, sizeof(formdata));
    }
    else if (strcasecmp(method, "GET") == 0)
    {
-      read_httpget_data(url, formdata);         
+      formdatalen = read_httpget_data(url, formdata, sizeof(formdata));
    }
    else
    {
@@ -435,15 +553,49 @@ int process_http_request(int sock)
    
    /* Open file for output */
    fout = fdopen(sock, "w");
+
+   /* Send an error if we failed to read the form data properly */
+   if (formdatalen < 0) {
+     send_headers(fout, 500, "Internal Error", NULL, "text/html", -1, -1);
+     fprintf(fout, "ERROR: Invalid Input. \r\n");
+     goto done;
+   }
+
+   /* Get values from form data */
+   if (formdatalen > 0) 
+   {
+      int found=0;
+      //printf("DBG: Found data: ");
+      datastr = strstr(formdata, RELAY_TAG);
+      if (datastr) {
+         relay = atoi(datastr+strlen(RELAY_TAG)+1);
+         //printf("relay=%d", relay);
+         found++;
+      }
+      datastr = strstr(formdata, STATE_TAG);
+      if (datastr) {
+         nstate = atoi(datastr+strlen(STATE_TAG)+1);
+         //printf("%sstate=%d", found ? ", " : "", nstate);
+      }
+      datastr = strstr(formdata, SERIAL_TAG);
+      if (datastr) {
+         serial = datastr+strlen(SERIAL_TAG)+1;
+         datastr = strstr(serial, "&");
+         if (datastr)
+            *datastr = 0;
+         //printf("%sserial=%s", found ? ", " : "", serial);
+      }
+      //printf("\n\n");
+   }
    
    /* Check if a relay card is present */
-   if (detect_relay_card(com_port, &last_relay) == -1)
+   if (crelay_detect_relay_card(com_port, &last_relay, serial, NULL) == -1)
    {
       if (strstr(url, API_URL))
       {
          /* HTTP API request, send response */
-         send_headers(fout, 200, "OK", NULL, "text/html", -1, -1);
-         fprintf(fout, "ERROR: No compatible device detected !\r\n");
+         send_headers(fout, 503, "No compatible device detected", NULL, "text/plain", -1, -1);
+         fprintf(fout, "ERROR: No compatible device detected");
       }
       else
       {  
@@ -456,42 +608,32 @@ int process_http_request(int sock)
    else
    {  
       /* Process form data */
-      if (strlen(formdata)>0)
+      if (formdatalen > 0)
       {
-         /* Get values from form data */
-         datastr = strstr(formdata, RELAY_TAG);
-         if (datastr)
-            relay = atoi(datastr+strlen(RELAY_TAG)+1);
-         datastr = strstr(formdata, STATE_TAG);
-         if (datastr)
-            nstate = atoi(datastr+strlen(STATE_TAG)+1);
-         
-         //printf("DBG: Found data: relay=%d, state=%d\n\n", relay, nstate);
-         
          if ((relay != 0) && (nstate != INVALID))
          {
             /* Perform the requested action here */
             if (nstate==PULSE)
             {
                /* Generate pulse on relay switch */
-               get_relay(com_port, relay, &rstate[relay-1]);
+               crelay_get_relay(com_port, relay, &rstate[relay-1], serial);
                if (rstate[relay-1] == ON)
                {
-                  set_relay(com_port, relay, OFF);
-                  sleep(1);
-                  set_relay(com_port, relay, ON);
+                  crelay_set_relay(com_port, relay, OFF, serial);
+                  sleep(config.pulse_duration);
+                  crelay_set_relay(com_port, relay, ON, serial);
                }
                else
                {
-                  set_relay(com_port, relay, ON);
-                  sleep(1);
-                  set_relay(com_port, relay, OFF);
+                  crelay_set_relay(com_port, relay, ON, serial);
+                  sleep(config.pulse_duration);
+                  crelay_set_relay(com_port, relay, OFF, serial);
                }
             }
             else
             {
                /* Switch relay on/off */
-               set_relay(com_port, relay, nstate);
+               crelay_set_relay(com_port, relay, nstate, serial);
             }
          }
       }
@@ -499,14 +641,14 @@ int process_http_request(int sock)
       /* Read current state for all relays */
       for (i=FIRST_RELAY; i<=last_relay; i++)
       {
-         get_relay(com_port, i, &rstate[i-1]);
+         crelay_get_relay(com_port, i, &rstate[i-1], serial);
       }
       
       /* Send response to client */
       if (strstr(url, API_URL))
       {
          /* HTTP API request, send response */
-         send_headers(fout, 200, "OK", NULL, "text/html", -1, -1);
+         send_headers(fout, 200, "OK", NULL, "text/plain", -1, -1);
          for (i=FIRST_RELAY; i<=last_relay; i++)
          {
             fprintf(fout, "Relay %d:%d<br>", i, rstate[i-1]);
@@ -516,13 +658,12 @@ int process_http_request(int sock)
       {
          /* Web request */
          char cname[MAX_RELAY_CARD_NAME_LEN];
-         get_relay_card_name(get_relay_card_type(), cname);
+         crelay_get_relay_card_name(crelay_get_relay_card_type(), cname);
          
          web_page_header(fout);
          
          /* Display relay status and controls on web page */
-         fprintf(fout, "<img style=\"width: 250px; height: 250px;\" alt=\"Card image\" src=\"http://www.conrad.de/medias/global/ce/1000_1999/1900/1920/1928/192846_LB_00_FB.EPS_1000.jpg\">\r\n");
-         fprintf(fout, "<table style=\"text-align: left; width: 600px; background-color: white; font-family: Helvetica,Arial,sans-serif; font-weight: bold; font-size: 20px;\" border=\"0\" cellpadding=\"2\" cellspacing=\"3\"><tbody>\r\n");
+         fprintf(fout, "<table style=\"text-align: left; width: 460px; background-color: white; font-family: Helvetica,Arial,sans-serif; font-weight: bold; font-size: 20px;\" border=\"0\" cellpadding=\"2\" cellspacing=\"3\"><tbody>\r\n");
          fprintf(fout, "<tr style=\"font-size: 14px; background-color: lightgrey\">\r\n");
          fprintf(fout, "<td style=\"width: 200px;\">%s<br><span style=\"font-style: italic; font-size: 12px; color: grey; font-weight: normal;\">on %s</span></td>\r\n", 
                  cname, com_port);
@@ -532,19 +673,17 @@ int process_http_request(int sock)
             fprintf(fout, "<tr style=\"vertical-align: top; background-color: rgb(230, 230, 255);\">\r\n");
             fprintf(fout, "<td style=\"width: 300px;\">Relay %d<br><span style=\"font-style: italic; font-size: 16px; color: grey;\">%s</span></td>\r\n", 
                     i, rlabels[i-1]);
-            fprintf(fout, "<td style=\"text-align: center; vertical-align: middle; width: 100px; background-color: %s;\">%s</td>\r\n", 
-                    rstate[i-1]==ON?"red":"lightgrey", rstate[i-1]==ON?"ON":"OFF");
-            fprintf(fout, "<td style=\"text-align: center; vertical-align: middle; width: 100px;\"><form action='/' method='post'><input type='hidden' name='%s' value='%d' /><input type='hidden' name='%s' value='%d' /><input type='submit' title='Toggle relay' value='%s'></form></td>\r\n", 
-                    RELAY_TAG, i, STATE_TAG, rstate[i-1]==ON?0:1, rstate[i-1]==ON?"off":"on");
-            fprintf(fout, "<td style=\"text-align: center; vertical-align: middle; width: 100px;\"><form action='/' method='post'><input type='hidden' name='%s' value='%d' /><input type='hidden' name='%s' value='2' /><input type='submit' title='Generate pulse' value='pulse'></form></td></tr>\r\n", 
-                    RELAY_TAG, i, STATE_TAG);
+            fprintf(fout, "<td style=\"text-align: center; vertical-align: middle; width: 100px; background-color: white;\"><label class=\"switch\"><input type=\"checkbox\" %s id=%d onchange=\"switch_relay(this)\"><span class=\"slider\"></span></label></td>\r\n", 
+                    rstate[i-1]==ON?"checked":"",i);
          }
          fprintf(fout, "</tbody></table><br>\r\n");
+         fprintf(fout, "<span id=\"status\" style=\"font-size: 16px; color: red; font-family: Helvetica,Arial,sans-serif;\"></span><br><br>\r\n");
          
          web_page_footer(fout);
       }
    }
-   
+
+ done:
    fclose(fout);
    fclose(fin);
 
@@ -570,21 +709,20 @@ void print_usage()
    printf("Supported relay cards:\n");
    for(rtype=NO_RELAY_TYPE+1; rtype<LAST_RELAY_TYPE; rtype++)
    {
-      get_relay_card_name(rtype, cname);
+      crelay_get_relay_card_name(rtype, cname);
       printf("  - %s\n", cname);
    }
-   printf("The card which is detected first will be used.\n");
    printf("\n");
    printf("The program can be run in interactive (command line) mode or in daemon mode with\n");
    printf("built-in web server.\n\n");
    printf("Interactive mode:\n");
-   printf("    crelay -i | [<relay number>] [ON|OFF]\n\n");
+   printf("    crelay -i | [-s <serial number>] <relay number> [ON|OFF]\n\n");
    printf("       -i print relay information\n\n");
    printf("       The state of any relay can be read or it can be changed to a new state.\n");
    printf("       If only the relay number is provided then the current state is returned,\n");
    printf("       otherwise the relays state is set to the new value provided as second parameter.\n");
    printf("       The USB communication port is auto detected. The first compatible device\n");
-   printf("       found will be used.\n\n");
+   printf("       found will be used, unless -s switch and a serial number is passed.\n\n");
    printf("Daemon mode:\n");
    printf("    crelay -d|-D [<relay1_label> [<relay2_label> [<relay3_label> [<relay4_label>]]]] \n\n");
    printf("       -d use daemon mode, run in foreground\n");
@@ -612,6 +750,16 @@ void print_usage()
  *********************************************************/
 int main(int argc, char *argv[])
 {
+      relay_state_t rstate;
+      char com_port[MAX_COM_PORT_NAME_LEN];
+      char cname[MAX_RELAY_CARD_NAME_LEN];
+      char* serial=NULL;
+      uint8_t num_relays=FIRST_RELAY;
+      relay_info_t *relay_info;
+      int argn = 1;
+      int err;
+      int i = 1;
+
    if (argc==1)
    {
       print_usage();
@@ -629,7 +777,7 @@ int main(int argc, char *argv[])
       int i;
       
       iface.s_addr = INADDR_ANY;
-      
+
       
       openlog("crelay", LOG_PID|LOG_CONS, LOG_USER);
       syslog(LOG_DAEMON | LOG_NOTICE, "Starting crelay daemon (version %s)\n", VERSION);
@@ -654,7 +802,9 @@ int main(int argc, char *argv[])
          if (config.relay6_label != NULL) syslog(LOG_DAEMON | LOG_NOTICE, "relay6_label: %s\n", config.relay6_label);
          if (config.relay7_label != NULL) syslog(LOG_DAEMON | LOG_NOTICE, "relay7_label: %s\n", config.relay7_label);
          if (config.relay8_label != NULL) syslog(LOG_DAEMON | LOG_NOTICE, "relay8_label: %s\n", config.relay8_label);
+         if (config.pulse_duration != 0)  syslog(LOG_DAEMON | LOG_NOTICE, "pulse_duration: %u\n", config.pulse_duration);
          if (config.gpio_num_relays != 0) syslog(LOG_DAEMON | LOG_NOTICE, "gpio_num_relays: %u\n", config.gpio_num_relays);
+         if (config.gpio_active_value >= 0) syslog(LOG_DAEMON | LOG_NOTICE, "gpio_active_value: %u\n", config.gpio_active_value);
          if (config.relay1_gpio_pin != 0) syslog(LOG_DAEMON | LOG_NOTICE, "relay1_gpio_pin: %u\n", config.relay1_gpio_pin);
          if (config.relay2_gpio_pin != 0) syslog(LOG_DAEMON | LOG_NOTICE, "relay2_gpio_pin: %u\n", config.relay2_gpio_pin);
          if (config.relay3_gpio_pin != 0) syslog(LOG_DAEMON | LOG_NOTICE, "relay3_gpio_pin: %u\n", config.relay3_gpio_pin);
@@ -690,11 +840,18 @@ int main(int argc, char *argv[])
          {
             port = config.server_port;
          }
+
       }
       else
       {
          syslog(LOG_DAEMON | LOG_NOTICE, "Can't load %s, using default parameters\n", CONFIG_FILE);
       }
+
+      /* Ensure pulse duration is valid **/
+      if (config.pulse_duration == 0)
+      {
+         config.pulse_duration = 1;
+      }
       
       /* Parse command line for relay labels (overrides config file)*/
       for (i=0; i<argc-2 && i<MAX_NUM_RELAYS; i++)
@@ -707,8 +864,17 @@ int main(int argc, char *argv[])
       sin.sin_family = AF_INET;
       sin.sin_addr.s_addr = iface.s_addr;
       sin.sin_port = htons(port);
-      bind(sock, (struct sockaddr *) &sin, sizeof(sin));      
-      listen(sock, 5);
+      if (bind(sock, (struct sockaddr *) &sin, sizeof(sin)) != 0)
+      {
+         syslog(LOG_DAEMON | LOG_ERR, "Failed to bind socket to port %d : %s", port, strerror(errno));
+         exit(EXIT_FAILURE);         
+      }
+      if (listen(sock, 5) != 0)
+      {
+         syslog(LOG_DAEMON | LOG_ERR, "Failed to listen to port %d : %s", port, strerror(errno));
+         exit(EXIT_FAILURE);         
+      }
+      
       syslog(LOG_DAEMON | LOG_NOTICE, "HTTP server listening on %s:%d\n", inet_ntoa(iface), port);      
 
       if (!strcmp(argv[1],"-D"))
@@ -722,6 +888,9 @@ int main(int argc, char *argv[])
          syslog(LOG_DAEMON | LOG_NOTICE, "Program is now running as system daemon");
       }
 
+      /* Init GPIO pins in case they have been configured */
+      crelay_detect_relay_card(com_port, &num_relays, NULL, NULL);
+      
       while (1)
       {
          int s;
@@ -741,49 +910,73 @@ int main(int argc, char *argv[])
    {
       /*****  Command line mode *****/
       
-      relay_state_t rstate;
-      char com_port[MAX_COM_PORT_NAME_LEN];
-      char cname[MAX_RELAY_CARD_NAME_LEN];
-      uint8 num_relays=FIRST_RELAY;
-      int err;
+      if (!strcmp(argv[argn],"-i"))
+      {
+         /* Detect all cards connected to the system */
+         if (crelay_detect_all_relay_cards(&relay_info) == -1)
+         {
+            printf("No compatible device detected.\n");
+            return -1;
+         }
+         printf("\nDetected relay cards:\n");
+         while (relay_info->next != NULL)
+         {
+            crelay_get_relay_card_name(relay_info->relay_type, cname);
+            printf("  #%d\t%s (serial %s)\n", i++ ,cname, relay_info->serial);
+            relay_info = relay_info->next;
+            // TODO: free relay_info list memory
+         }
+         
+         exit(EXIT_SUCCESS);
+      }
+   
+      if (!strcmp(argv[argn], "-s"))
+      {
+         if (argc > (argn+1))
+         {
+            serial = malloc(sizeof(char)*strlen(argv[argn+1]));
+            strcpy(serial, argv[argn+1]);
+            argn += 2;
+         }
+         else
+         {
+            print_usage();
+            exit(EXIT_FAILURE);            
+         }
+      }
 
-      if (detect_relay_card(com_port, &num_relays) == -1)
+      if (crelay_detect_relay_card(com_port, &num_relays, serial, NULL) == -1)
       {
          printf("No compatible device detected.\n");
          
          if(geteuid() != 0)
          {
-            printf("\nWarning: this program is currently not running with root priviledges !\n");
-            printf("Therefore it might not be able to access your relay cards communication port.\n");
+            printf("\nWarning: this program is currently not running with root privileges !\n");
+            printf("Therefore it might not be able to access your relay card communication port.\n");
             printf("Consider invoking the program from the root account or use \"sudo ...\"\n");
          }
          
          exit(EXIT_FAILURE);
       }
 
-      if (!strcmp(argv[1],"-i"))
-      {
-         if (get_relay_card_name(get_relay_card_type(), cname) == 0)
-            printf("Detected relay card type is %s (on %s, %d channels)\n", cname, com_port, num_relays);
-         exit(EXIT_SUCCESS);
-      }
-   
       switch (argc)
       {
          case 2:
+         case 4:
             /* GET current relay state */
-            if (get_relay(com_port, atoi(argv[1]), &rstate) == 0)
-               printf("Relay %d is %s\n", atoi(argv[1]), (rstate==ON)?"on":"off");
+            if (crelay_get_relay(com_port, atoi(argv[argn]), &rstate, serial) == 0)
+               printf("Relay %d is %s\n", atoi(argv[argn]), (rstate==ON)?"on":"off");
             else
                exit(EXIT_FAILURE);
             break;
             
          case 3:
+         case 5:
             /* SET new relay state */
-            if (!strcmp(argv[2],"on") || !strcmp(argv[2],"ON")) 
-               err = set_relay(com_port, atoi(argv[1]), ON);
-            else if (!strcmp(argv[2],"off") || !strcmp(argv[2],"OFF")) 
-               err = set_relay(com_port, atoi(argv[1]), OFF);
+            if (!strcmp(argv[argn+1],"on") || !strcmp(argv[argn+1],"ON"))
+               err = crelay_set_relay(com_port, atoi(argv[argn]), ON, serial);
+            else if (!strcmp(argv[argn+1],"off") || !strcmp(argv[argn+1],"OFF"))
+               err = crelay_set_relay(com_port, atoi(argv[argn]), OFF, serial);
             else 
             {
                print_usage();
@@ -797,5 +990,6 @@ int main(int argc, char *argv[])
             print_usage();
       }
    }
+   
    exit(EXIT_SUCCESS);
 }
